import { derived, get, writable, type Writable } from "svelte/store"
import { API_BASE_ADDR, AT_EXP_MS, CSRF_EXP_MS, REFRESH_BUF, UUID_REGEX } from "./const"
import { goto } from "$app/navigation"
import { usersPagination } from "./pages"

interface UserResponse {
	id: string
	username: string
	isAdmin: boolean
	createdAt: Date
	updatedAt: Date
}

export const currentUser: Writable<UserResponse | null> = writable(null)
export const accessToken: Writable<string | null> = writable(null)
export const csrfToken: Writable<string | null> = writable(null)
export const isLoading: Writable<boolean> = writable(false)
export const cError: Writable<string | null> = writable(null)

class ApiClient {
	private baseUrl: string
	private lastAtUpdate = new Date(0)
	private lastCsrfUpdate = new Date(0)
	private refreshInProgress = false

	constructor(baseUrl: string) {
		this.baseUrl = baseUrl
	}

	private async handleRequest<T>(
		fn: () => Promise<T>,
		options: { useBearerAuth: boolean }
	): Promise<T | undefined> {
		isLoading.set(true)
		cError.set(null)

		// If `handleResponse` is used, errors thrown from it will be caught here

		try {
			// Only bearers (JWT access tokens) need to be considered here as the refresh token cookies
			// will be automatically included to corresponding requests by browser and CSRF rotation is
			// handled by the `refreshAccessToken` method
			if (options.useBearerAuth) {
				await this.checkAndRefreshAccessToken()
			}
			return await fn()
		} catch (err) {
			cError.set(err instanceof Error ? err.message : "Unknown error")
			console.log(`error: ${get(cError)}`)
		} finally {
			isLoading.set(false)
		}
	}

	// Should be attached to routes that handle authentication with the bearer token (access token)
	private async handleResponse<T>(response: Response): Promise<T> {
		if (!response.ok) {
			if (response.status === 401) {
				// This should never happen due to the token expiration checks we make client-side,
				// but it's still good to have as a fallback
				try {
					console.log("unexpected 401 caught, attempting refresh")
					await this.refreshAccessToken()
				} catch (err) {
					console.log("refresh attempt not successful")
					await this.handleLocalLogout()
					throw new Error("Session expired, please authenticate again.")
				}
			}

			// Capitalize the error message and display (only if not 401)
			const { error } = await response.json()
			console.log(`response error: ${error}`)
			const dError = error[0].toUpperCase() + error.substr(1).toLowerCase() + "."

			throw new Error(dError)
		}

		return response.json()
	}

	private async checkAndRefreshAccessToken(): Promise<void> {
		if (this.refreshInProgress) return

		const timeSinceUpdate = Date.now() - this.lastAtUpdate.getTime()
		const needsRefresh = timeSinceUpdate > AT_EXP_MS - REFRESH_BUF

		console.log(`timeSinceUpdate: ${timeSinceUpdate}`)

		if (needsRefresh) {
			console.log("running token refresh attempt")

			this.refreshInProgress = true
			await this.refreshAccessToken()
			this.refreshInProgress = false
		}
	}

	private async refreshAccessToken(): Promise<void> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/cookie/refresh`, {
					method: "POST",
					credentials: "include",
					headers: {
						...(await this.getCsrfHeader())
					}
				})

				if (!response.ok) {
					await this.handleLocalLogout()
					throw new Error("Refreshing token failed.")
				}

				const { accessToken: newToken } = await response.json()
				accessToken.set(newToken)
				this.lastAtUpdate = new Date()
			},
			{ useBearerAuth: false }
		)
	}

	private async getCsrfHeader(): Promise<HeadersInit> {
		let token = get(csrfToken)
		const timeSinceUpdate = Date.now() - this.lastCsrfUpdate.getTime()
		const needsRefresh = timeSinceUpdate > CSRF_EXP_MS - REFRESH_BUF

		if (!token || needsRefresh) {
			await this.refreshCsrfToken()
			token = get(csrfToken)
		}

		return { "X-CSRF-Token": token || "" }
	}

	private async refreshCsrfToken(): Promise<void> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/cookie/csrf`, {
					credentials: "include"
				})

				if (!response.ok) {
					await this.handleLocalLogout()
					throw new Error("Fetching CSRF token failed.")
				}

				const newToken = response.headers.get("X-CSRF-Token")
				csrfToken.set(newToken)
				this.lastCsrfUpdate = new Date()
			},
			{ useBearerAuth: false }
		)
	}

	private async handleLocalLogout(): Promise<void> {
		this.lastAtUpdate = new Date(0)
		this.lastCsrfUpdate = new Date(0)
		currentUser.set(null)
		accessToken.set(null)
		csrfToken.set(null)
		await goto("/login")
	}

	private getAuthHeader(): HeadersInit {
		const token = get(accessToken)
		return { Authorization: `Bearer ${token}` }
	}

	public async register(username: string, password: string): Promise<void> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/signup`, {
					method: "POST",
					headers: { "Content-Type": "application/json" },
					body: JSON.stringify({ username, password })
				})

				// Can't overwrite `username` parameter
				const data = await this.handleResponse<{
					id: string
					username: string
				}>(response)

				console.log(`'${data.username} -> ${data.id}'`)
				await goto("/login")
			},
			{ useBearerAuth: false }
		)
	}

	public async login(username: string, password: string): Promise<void> {
		return this.handleRequest(
			async () => {
				const params = new URLSearchParams()
				params.append("includeUser", "true")
				const response = await fetch(`${this.baseUrl}/auth/login?${params}`, {
					method: "POST",
					headers: { "Content-Type": "application/json" },
					body: JSON.stringify({ username, password })
				})

				const { accessToken: token, user } = await this.handleResponse<{
					accessToken: string
					user: UserResponse
				}>(response)

				accessToken.set(token)
				currentUser.set(user || null)
				this.lastAtUpdate = new Date()

				console.log(user)
			},
			{ useBearerAuth: false }
		)
	}

	public async logout(): Promise<void> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/logout`, {
					method: "POST",
					headers: {
						...this.getAuthHeader()
					}
				})

				if (response.status === 204) {
					console.log("logout successful")
					await this.handleLocalLogout()
					return
				}

				await this.handleResponse<void>(response)
				await this.handleLocalLogout()
			},
			{ useBearerAuth: true }
		)
	}

	public async getCurrentUser(): Promise<void> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/me`, {
					headers: {
						...this.getAuthHeader()
					}
				})
				const user = await this.handleResponse<UserResponse>(response)
				currentUser.set(user)
			},
			{ useBearerAuth: true }
		)
	}

	public async updateCurrentUserPassword(oldPassword: string, newPassword: string): Promise<void> {
		return this.handleRequest(
			async () => {
				const data = {
					old_password: oldPassword,
					new_password: newPassword
				}
				const response = await fetch(`${this.baseUrl}/auth/owner`, {
					method: "PUT",
					headers: {
						...this.getAuthHeader()
					},
					body: JSON.stringify(data)
				})
				const { accessToken: token, user } = await this.handleResponse<{
					accessToken: string
					user: UserResponse
				}>(response)
				accessToken.set(token)
				currentUser.set(user || null)
				this.lastAtUpdate = new Date()

				console.log(user)
			},
			{ useBearerAuth: true }
		)
	}

	public async deleteCurrentUser(password: string): Promise<void> {
		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/owner`, {
					method: "DELETE",
					headers: {
						...this.getAuthHeader()
					},
					body: JSON.stringify({ password })
				})

				if (response.status === 204) {
					console.log("deletion succesful")
					await this.handleLocalLogout()
					return
				}

				await this.handleResponse<void>(response)
				await this.handleLocalLogout()
			},
			{ useBearerAuth: true }
		)
	}

	public async adminListAll(): Promise<UserResponse[] | undefined> {
		const user = get(currentUser)
		if (!user || !user.isAdmin) {
			throw new Error("Admin privileges required.")
		}

		return this.handleRequest(
			async () => {
				const params = new URLSearchParams()
				params.append("limit", `${get(usersPagination).pageSize}`)
				params.append("offset", `${get(usersPagination).currentPage}`)
				const response = await fetch(`${this.baseUrl}/auth/admin/all`, {
					headers: {
						...this.getAuthHeader()
					}
				})

				const users = await this.handleResponse<UserResponse[]>(response)
				console.log(`admin: got ${users.length} results`)

				return users
			},
			{ useBearerAuth: true }
		)
	}

	public async adminDeleteUser(userID: string): Promise<void> {
		const user = get(currentUser)
		if (!user || !user.isAdmin) {
			throw new Error("Admin privileges required.")
		}

		if (!UUID_REGEX.test(userID)) {
			throw new Error("Invalid user ID format.")
		}

		return this.handleRequest(
			async () => {
				const response = await fetch(`${this.baseUrl}/auth/admin/${userID}`, {
					method: "DELETE",
					headers: {
						...this.getAuthHeader()
					}
				})

				if (response.status === 204) {
					console.log("admin: deletion successful")
					return
				}

				await this.handleResponse<void>(response)
			},
			{ useBearerAuth: true }
		)
	}
}

export const apiClient = new ApiClient(API_BASE_ADDR)
